import gymnasium as gym
from gymnasium import spaces
from gymnasium import utils
import numpy as np
import matplotlib.pyplot as plt

import logging
logger = logging.getLogger(__name__)


class InventoryEnv(gym.Env, utils.EzPickle):
    """Inventory control with lost sales environment
    TO BE EDITED
    This environment corresponds to the version of the inventory control
    with lost sales problem described in Example 1.1 in Algorithms for
    Reinforcement Learning by Csaba Szepesvari (2010).
    https://sites.ualberta.ca/~szepesva/RLBook.html
    """
    def __init__(self, n=100, k=5, c=2, h=2, p=3, lam=8):
        self.n = n
        self.action_space = spaces.Discrete(n)
        self.observation_space = spaces.Discrete(n)
        self.max = n
        self.state = n
        self.k = k
        self.c = c
        self.h = h
        self.p = p
        self.lam = lam
    def demand(self):
        return np.random.poisson(self.lam)
    def plot_demand(self):
        data = np.random.poisson(self.lam, 1000)
        # Plot the histogram
        plt.hist(data, bins=range(data.min(), data.max()+2), align='left')
        plt.xlabel('Value')
        plt.ylabel('Frequency')
        plt.title('Histogram of Demand Distribution')
        plt.show()
    def transition(self, x, a, d):
        m = self.max
        return max(min(x + a, m) - d, 0)
    def reward(self, x, a, y):
        k = self.k
        m = self.max
        c = self.c
        h = self.h
        p = self.p
        r = -k * (a > 0) - c * max(min(x + a, m) - x, 0) - h * x + p * max(min(x + a, m) - y, 0)
        return r
    def step(self, action):
        assert self.action_space.contains(action)
        obs = self.state
        demand = self.demand()
        obs2 = self.transition(obs, action, demand)
        self.state = obs2
        reward = self.reward(obs, action, obs2)
        terminated = True
        info = {}
        truncated = False
        return obs2, reward, terminated, truncated, info
    def reset(self, seed=None, options=None):
        super().reset(seed=seed)
        observation = self.observation_space.sample()
        self.state = observation
        info = {}
        #print(observation)
        return self.state,info  # reward, done, info can't be included
if __name__ == '__main__':
    a=InventoryEnv()
    a.plot_demand()


import gymnasium as gym
from gymnasium import spaces
import numpy as np
class AsymmetricInfoEnv(gym.Env):
    metadata = {"render.modes": ["human"]}
    def __init__(self):
        self.observation_space = spaces.Discrete(2)
        self.action_space = spaces.Discrete(2)
        self.reward_range = (-1, 1)
        self.additional_info = None
    def transition(self):
        pass
    def reward(self, x, a, y):
        pass
    def step(self, action):
        observation = self.observation_space.sample()
        if action == 0:
            if self.additional_info is None:
                reward = -1
            else:
                reward = 1
        else:
            reward = 0
        terminated = False
        info = {}
        truncated = False
        return observation, reward, terminated, truncated, info
    def reset(self, seed=None, options=None):
        super().reset(seed=seed)
        self.additional_info = None
        observation = self.observation_space.sample()
        info = {}
        return observation,info
    def provide_additional_info(self, info):
        self.additional_info = info
    def render(self, mode="human"):
        pass
    def close(self):
        pass
if __name__ == '__main__':
    a=AsymmetricInfoEnv()
    print(a.action_space)


from cpagym.envs.inventory_env import *
from cpagym.envs.trading_env import *
from cpagym.envs.forex_env import *
from cpagym.envs.stocks_env import *
from cpagym.envs.moral_hazard_env import *
from cpagym.envs.asymmetricInfo_env import *
from cpagym.train.DQN import *


import gymnasium as gym
import torch
import torch.nn as nn
import torch.optim as optim
import numpy as np
from collections import namedtuple, deque
import random
import math
import matplotlib
import matplotlib.pyplot as plt
from itertools import count
from cpagym.envs.moral_hazard_env import *

class DQN(nn.Module):
    def __init__(self, input_size, hidden_size, output_size):
        super(DQN, self).__init__()
        self.fc1 = nn.Linear(input_size, hidden_size)
        self.fc2 = nn.Linear(hidden_size, output_size)

    def forward(self, x):
        x = torch.relu(self.fc1(x))
        x = self.fc2(x)
        return x
def train_dqn(env, model, optimizer, num_episodes, epsilon_decay):
    for episode in range(num_episodes):
        state = env.reset()
        done = False
        while not done:
            q_values = torch.zeros((1,2), dtype=torch.float32)
            # Choose an action using epsilon-greedy exploration
            if random.random() < epsilon_decay:
                action = env.action_space.sample()
            else:
                state_tensor = torch.tensor([state], dtype=torch.float32)
                q_values = model(state_tensor)
                action = q_values.argmax().item()
            next_state, reward, done,_, _ = env.step(action)
            # Update the model using the Bellman equation
            next_state_tensor = torch.tensor([next_state], dtype=torch.float32)
            q_values_next = model(next_state_tensor)
            q_target = reward + 0.99 * q_values_next.max()
            q_values[0, action] = q_target
            loss = ((q_values - q_values_next) ** 2).mean()
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            state = next_state

def main():
    env = Moral_HazardEnv()
    model = DQN(1, 128, 2)
    optimizer = optim.SGD(model.parameters(), lr=0.01)
    num_episodes = 1000
    epsilon_decay = 1.0
    epsilon_min = 0.01
    for episode in range(num_episodes):
        epsilon_decay = max(epsilon_min, epsilon_decay * 0.999)
        train_dqn(env, model, optimizer, 1, epsilon_decay)
    epsilon_decay = 0
    total_reward = 0
    for episode in range(100):
        state = env.reset()
        done = False
        while not done:
            state_tensor = torch.tensor([state], dtype=torch.float32)
            q_values = model(state_tensor)
            action = q_values.argmax().item()
            next_state, reward, done,_, _ = env.step(action)

            total_reward += reward
            state = next_state
    
    # Print the average reward over 100 episodes
    print(f"Average reward: {total_reward / 100}")
if __name__ == '__main__':
    main()